<?php

declare(strict_types=1);

namespace Symplify\PHPStanRules\Rules;

use PhpParser\Node;
use PhpParser\Node\Stmt\Class_;
use PHPStan\Analyser\Scope;
use PHPStan\Node\InClassNode;
use PHPStan\Rules\Rule;
use PHPStan\Rules\RuleError;
use PHPStan\Rules\RuleErrorBuilder;
use Symplify\PHPStanRules\NodeAnalyzer\ProtectedAnalyzer;
use Symplify\PHPStanRules\ParentGuard\ParentPropertyGuard;
use Symplify\RuleDocGenerator\Contract\DocumentedRuleInterface;
use Symplify\RuleDocGenerator\ValueObject\CodeSample\CodeSample;
use Symplify\RuleDocGenerator\ValueObject\RuleDefinition;

/**
 * @see \Symplify\PHPStanRules\Tests\Rules\ForbiddenProtectedPropertyRule\ForbiddenProtectedPropertyRuleTest
 *
 * @implements Rule<InClassNode>
 */
final class ForbiddenProtectedPropertyRule implements Rule
{
    /**
     * @var string
     */
    public const ERROR_MESSAGE = 'Property with protected modifier is not allowed. Use interface contract method instead';
    /**
     * @var \Symplify\PHPStanRules\NodeAnalyzer\ProtectedAnalyzer
     */
    private $protectedAnalyzer;
    /**
     * @var \Symplify\PHPStanRules\ParentGuard\ParentPropertyGuard
     */
    private $parentPropertyGuard;
    public function __construct(ProtectedAnalyzer $protectedAnalyzer, ParentPropertyGuard $parentPropertyGuard)
    {
        $this->protectedAnalyzer = $protectedAnalyzer;
        $this->parentPropertyGuard = $parentPropertyGuard;
    }

    /**
     * @return class-string<InClassNode>
     */
    public function getNodeType(): string
    {
        return InClassNode::class;
    }

    /**
     * @param InClassNode $node
     * @return RuleError[]
     */
    public function processNode(Node $node, Scope $scope): array
    {
        $classLike = $node->getOriginalNode();
        if (! $classLike instanceof Class_) {
            return [];
        }

        $errorMessages = [];

        foreach ($classLike->getProperties() as $property) {
            if (! $property->isProtected()) {
                continue;
            }

            if ($this->parentPropertyGuard->isPropertyGuarded($property, $scope)) {
                continue;
            }

            if ($this->protectedAnalyzer->isProtectedPropertyOrClassConstAllowed($property, $classLike)) {
                continue;
            }

            $errorMessages[] = RuleErrorBuilder::message(self::ERROR_MESSAGE)
                ->line($property->getLine())
                ->build();
        }

        return $errorMessages;
    }

    public function getRuleDefinition(): RuleDefinition
    {
        return new RuleDefinition(self::ERROR_MESSAGE, [
            new CodeSample(
                <<<'CODE_SAMPLE'
class SomeClass
{
    protected $repository;
}
CODE_SAMPLE
                ,
                <<<'CODE_SAMPLE'
class SomeClass implements RepositoryAwareInterface
{
    public function getRepository()
    {
        // ....
    }
}
CODE_SAMPLE
            ),
        ]);
    }
}
